/* jshint undef: true, unused: true */
/* global plugin, require, module, java, Packages */

var facade = require("facade");
var _ = require("lodash");
var Objects = require("convenience").Objects;
var time = require("./time");

var STORE = plugin.info().dataStore;
var templateProxy = facade.retrieveProxy("CaptureTemplateProxy");

var JDate = java.util.Date;

var templates = {
    schedule: function schedule (args) {
        return _(plugin.resource("templates", "schedule.tmpl").read().getVal()).template()(args);
    },
    disabled: function disabled (args) {
        return _(plugin.resource("templates", "disabled.tmpl").read().getVal()).template()(args);
    }
};

var PREFIX = "/schedule";

/*
 * Default form config. This is also the template
 * for a fresh form config.
 */
var DEFAULT_FORM_CONFIG = {
    enabled: false,
    redirectUrl: "",
    title: {
        value: "",
        editable: true,
        queryParameter: true
    },
    template: {
        value: "",
        editable: true,
        queryParameter: true
    },
    startDate: {
        value: "",
        editable: true,
        queryParameter: true
    },
    startTime: {
        value: "",
        editable: true,
        queryParameter: true
    },
    duration: {
        value: "",
        editable: true,
        queryParameter: true
    }
};

/*
 * Default default values for fields, if there
 * is neither an admin default nor user value specified.
 */
var DEFAULT_DEFAULTS = {
    startDate: function () {
        return time.DATE_FORMAT.format(time.now());
    },
    startTime: function () {
        return time.TIME_FORMAT_24H.format(time.getNextHourOfToday());
    }
};

/*
 * Functions that can change certain fields into the format
 * the Custom Scheduler Form wants. For example, we can change
 * a timestamp into a date.
 */
var FIELD_FORMATTERS = {
    startDate: function (value) {
        // Handle legacy (v4) date format
        if (_(value + "").includes("/"))
        {
            return time.DATE_FORMAT.format(time.LEGACY_DATE_FORMAT.parse(value));
        }

        if (!_(value + "").includes("-"))
        {
            var timestamp = parseInt(value);
            if (!isNaN(timestamp))
            {
                return time.DATE_FORMAT.format(new JDate(timestamp));
            }
        }

        return value;
    },
    startTime: function (value) {
        if (!_(value + "").includes(":"))
        {
            var timestamp = parseInt(value);
            if (!isNaN(timestamp))
            {
                return time.TIME_FORMAT_24H.format(new JDate(timestamp));
            }
        }

        return value;
    },
    duration: function (value) {
        if (!_(value + "").includes(":"))
        {
            var durationMinutes = parseInt(value);
            if (!isNaN(durationMinutes))
            {
                return time.formatDurationMinutes(parseInt(value));
            }
        }

        return value;
    }
};

function getForms ()
{
    var forms = Objects.toJS(STORE.select("form:*"));
    if (!forms.length)
    {
        setFormConfig(DEFAULT_FORM_CONFIG);
        return [DEFAULT_FORM_CONFIG];
    }
    return forms;
}

function serveStatic (path, response)
{
    var file = plugin.resource(path);
    if (file.isFile())
    {
        response.header("Content-Type",file.type()).content(file.readBytes().getOrRethrow()).end();
        return true;
    }
}

function getFormConfig ()
{
    return _({}).extend(DEFAULT_FORM_CONFIG).extend(getForms()[0]).value();
}

function setFormConfig (config)
{
    STORE.set("form:scheduler", Objects.toJS(config, true));
    return getFormConfig();
}

function updateFormConfig (settings)
{
    return setFormConfig(_(getFormConfig()).extend(Objects.toJS(settings, true)).value());
}

function getFormParameters (request)
{
    return _(getFormConfig()).toPairs().map(function (pair) {
        var key = pair[0];
        var field = _({}).extend(pair[1]).value();

        if (field.queryParameter)
        {
            field.value = request.queryParam(key) || field.value;
        }

        if (key in DEFAULT_DEFAULTS)
        {
            field.value = field.value || DEFAULT_DEFAULTS[key]();
        }

        if (key in FIELD_FORMATTERS)
        {
            field.value = FIELD_FORMATTERS[key](field.value);
        }

        return [key, field];
    }).fromPairs().value();
}

function getFormValues (request)
{
    var parameters = getFormParameters(request);
    return _(parameters).toPairs().map(function (pair) {
        var key = pair[0];
        var parameter = pair[1];
        var value = parameter.value;

        if (parameter.editable)
        {
            value = request.postParam(key) || value;
        }

        return [key, value];
    }).fromPairs().value();
}

function serveForm (request, response)
{
    var helper = new Packages.cattura.web.server.vos.RequestResponseHelper(request, response);
    var config = getFormConfig();
    var isExternalScheduler = isExternalSchedulerEnabled();
    if (!config.enabled || isExternalScheduler)
    {
        response.content(templates.disabled({
            relativeRoot: helper.getRelativeRoot(),
            reason: isExternalScheduler ? "It cannot be used while an external scheduler is enabled." : "It must be enabled by admin in the Settings panel."
        })).end();
        return;
    }

    response.header("Content-Type", "text/html").content(templates.schedule({
        relativeRoot: helper.getRelativeRoot(),
        form: getFormParameters(request),
        templates: Objects.toJS(templateProxy.getTemplates()),
        error: request.queryParam("error"),
        eventId: request.queryParam("eventId")
    })).end();
}

function handleForm (request, response, control)
{
    /*
     * If a redirect url is specified in form config, prefer it
     * over the default (which is just the form url, with any success/error
     * messages prepended to it). The config-specified form url doesn't get modified
     * unless the user explicitly wants to receive those parameters.
     */
    var defaultRedirectUrl = buildDefaultRedirectUrl(request);
    var configRedirectUrl = getFormConfig().redirectUrl || "";

    var values = getFormValues(request);
    var extraData = parseExtraData(request);
    var metadata = extraData.metadata;
    delete extraData.metadata;
    var scheduleRequest = facade.sendRequest("CREATE_CAPTURE_EVENT", {
        presentationTitle: values.title,
        template: values.template,
        startTime: time.parseStartTimeParameters(values.startDate, values.startTime),
        duration: time.parseDurationParameter(values.duration),
        publishingOverrides: extraData,
        metadata: metadata
    });

    scheduleRequest.addListener(function () {
        var error;

        try
        {
            var result = scheduleRequest.get().getEvent();
            defaultRedirectUrl = appendQueryParameter(defaultRedirectUrl, "eventId", result.ID);
            configRedirectUrl = replaceUrlVariables(configRedirectUrl, result, request);
        }
        catch (exception)
        {
            error = getErrorMessage(exception);
        }

        /*
         * If there is an error, we want to redirect back to our form and not to the user's
         * redirect url.
         */
        if (error)
        {
            defaultRedirectUrl = appendQueryParameter(defaultRedirectUrl, "error", error);
            configRedirectUrl = "";
        }

        response.status(302).header("Location", configRedirectUrl || defaultRedirectUrl).end();
    }, control.handlerExecutor());
}

function appendQueryParameter (url, key, value)
{
    return url + (url.indexOf("?") == -1 ? "?" : "&") + key + "=" + value;
}

function stripQueryString (url)
{
    if (url.contains("?"))
    {
        url = url.substring(0, url.indexOf("?"));
    }
    return url;
}

/*
 * Build default redirect url for the request. We want to preserve all
 * query string parameters other than ones that indicate success/failure.
 */
function buildDefaultRedirectUrl (request)
{
    var url = stripQueryString(request.uri());
    _(Objects.toJS(request.queryParamKeys())).each(function (key) {
        if (key == "error" || key == "eventId")
        {
            return;
        }

        url = appendQueryParameter(url, key, request.queryParam(key));
    });
    return url;
}

function requestHandler (request, response, control)
{
    var path = stripQueryString(request.uri());
    if (!(path == PREFIX || path.startsWith(PREFIX + "/")))
    {
        control.nextHandler();
        return;
    }

    var helper = new Packages.cattura.web.server.vos.RequestResponseHelper(request, response);

    if (helper.redirectStripDoubleSlashes()) return;

    path = path.substring(PREFIX.length).replace("%20"," ");
    if (path.startsWith("/js/") || path.startsWith("/css/"))
    {
        if (serveStatic(path, response))
        {
            return;
        }
    }
    else if (path.isEmpty() || path == "/")
    {
        if (request.method().toLowerCase() == "post")
        {
            handleForm(request, response, control);
        }
        else
        {
            serveForm(request, response);
        }

        return;
    }

    control.nextHandler();
    return;
}

function getErrorMessage (exception)
{
    if (exception.javaException)
    {
        return getErrorMessage(exception.javaException);
    }
    else if (exception.cause)
    {
        return getErrorMessage(exception.cause);
    }

    return exception.message;
}

function parseExtraData (request)
{
    var extraData = {};
    _(Objects.toJS(request.queryParamKeys(), true)).each(function (key) {
        if (!_(key).includes("."))
        {
            return;
        }

        var parts = key.split(".");
        var parentKey = parts[0];
        var childKey = parts[1];

        if (!(parentKey in extraData))
        {
            extraData[parentKey] = {};
        }

        extraData[parentKey][childKey] = request.queryParam(key);
    });
    return extraData;
}

function replaceUrlVariables (url, event, request)
{
    url = url.replace(/\$eventId/g, event.ID)
             .replace(/\$title/g, event.title);

    _(Objects.toJS(request.queryParamKeys(), true)).each(function (key) {
        url = url.replace("$" + key, request.queryParam(key));
    });

    return url;
}

function isExternalSchedulerEnabled ()
{
    var scheduleProxy = facade.retrieveProxy("SchedulerProxy");
    return scheduleProxy.getScheduler().getEventSource().getSourceName() != "local";
}

module.exports = {
    getFormConfig: getFormConfig,
    setFormConfig: setFormConfig,
    updateFormConfig: updateFormConfig,
    requestHandler: requestHandler
};
